/*
    gcc -DYAC_FOR_CDO -Wall -g -O3 -c *.c
    g++ -DYAC_FOR_CDO -Drestrict= -Wall -g -O3 cdo_remap_conserv_test.cc cdo_grid.cc cdo_cell_search.cc cdo_remap_conserv.cc *.o
 */

#include "cdo_remap.h"

static void
cdo_generate_grid_reg2d(GridInfo &grid, double *coordinates_x, double *coordinates_y, size_t const (&num_cells)[2])
{
  generate_grid_reg2d_yac(grid, coordinates_x, coordinates_y, num_cells);
}   

static void test1()
{
  const double missval = -1.0;
  RemapSearch remapSearch;

  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  {
    // the global source grid is a 2x2 grid:
    //       06--10--07--11--08
    //       |       |       |
    //       06  02  08  03  09
    //       |       |       |
    //       03--05--04--07--05
    //       |       |       |
    //       01  00  03  01  04
    //       |       |       |
    //       00--00--01--02--02

    double coordinates_x[] = {-1,0,1};
    double coordinates_y[] = {-1,0,1};
    size_t const num_cells[2] = {2,2};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;

    cdo_generate_grid_reg2d(srcGrid, coordinates_x, coordinates_y, num_cells);
  }

  {
    double coordinates_x[] = {-0.5,0.5};
    double coordinates_y[] = {-0.5,0.5};
    size_t const num_cells[2] = {1,1};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;
    
    cdo_generate_grid_reg2d(tgtGrid, coordinates_x, coordinates_y, num_cells);
  }

  Varray<double> srcField(srcGrid.numCells);
  Varray<double> tgtField(tgtGrid.numCells);

  for (size_t i = 0; i < srcGrid.numCells; ++i) srcField[i] = i + 1;

  gridcell_search_create(remapSearch.gcs, srcGrid.numCells, srcGrid.numCorners, srcGrid.cellCornersLon, srcGrid.cellCornersLat);

  remap_conserv(NormOpt::DESTAREA, remapSearch, srcField, tgtField, missval);

  gridcell_search_delete(remapSearch.gcs);

  {
    double tgtFieldRef[] = {2.5};

    for (size_t i = 0; i < tgtGrid.numCells; ++i)
      {
        if (fabs(tgtFieldRef[i] - tgtField[i]) > 1e-3)
          printf("%s: wrong interpolation result\n", __func__);
      }
  }
}

static void test3()
{
  const double missval = -1.0;
  RemapSearch remapSearch;

  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  {
    // the global source grid is a 3x3 grid:
    //
    //       12--21--13--22--14--23--15
    //       |       |       |       |
    //       15  06  17  07  19  08  20
    //       |       |       |       |
    //       08--14--09--16--10--18--11
    //       |       |       |       |
    //       08  03  10  04  12  05  13
    //       |       |       |       |
    //       04--07--05--09--06--11--07
    //       |       |       |       |
    //       01  00  03  01  05  02  06
    //       |       |       |       |
    //       00--00--01--02--02--04--03

    double coordinates_x[] = {-1.5,-0.5,0.5,1.5};
    double coordinates_y[] = {-1.5,-0.5,0.5,1.5};
    size_t const num_cells[2] = {3,3};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;

    cdo_generate_grid_reg2d(srcGrid, coordinates_x, coordinates_y, num_cells);
  }

  {
    // the global target grid is a 2x2 grid:
    //
    //       06--10--07--11--08
    //       |       |       |
    //       06  02  08  03  09
    //       |       |       |
    //       03--05--04--07--05
    //       |       |       |
    //       01  00  03  01  04
    //       |       |       |
    //       00--00--01--02--02

    double coordinates_x[] = {-1,0,1};
    double coordinates_y[] = {-1,0,1};
    size_t const num_cells[2] = {2,2};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;
    
    cdo_generate_grid_reg2d(tgtGrid, coordinates_x, coordinates_y, num_cells);
  }

  Varray<double> srcField(srcGrid.numCells);
  Varray<double> tgtField(tgtGrid.numCells);

  for (size_t i = 0; i < srcGrid.numCells; ++i) srcField[i] = i + 1;

  gridcell_search_create(remapSearch.gcs, srcGrid.numCells, srcGrid.numCorners, srcGrid.cellCornersLon, srcGrid.cellCornersLat);

  remap_conserv(NormOpt::DESTAREA, remapSearch, srcField, tgtField, missval);

  gridcell_search_delete(remapSearch.gcs);

  {
    double tgtFieldRef[] = {3,4,6,7};

    for (size_t i = 0; i < tgtGrid.numCells; ++i)
      {
        if (fabs(tgtFieldRef[i] - tgtField[i]) > 1e-3)
          printf("%s: wrong interpolation result\n", __func__);
      }
  }
}

static void test4()
{
  const double missval = -1.0;
  RemapSearch remapSearch;

  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  {
    // the global source grid is a 4x4 grid:
    //
    //       20--36--21--37--22--38--23--39--24
    //       |       |       |       |       |
    //       28  12  30  13  32  14  34  15  35
    //       |       |       |       |       |
    //       15--27--16--29--17--31--18--33--19
    //       |       |       |       |       |
    //       19  08  21  09  23  10  25  11  26
    //       |       |       |       |       |
    //       10--18--11--20--12--22--13--24--14
    //       |       |       |       |       |
    //       10  04  12  05  14  06  16  07  17
    //       |       |       |       |       |
    //       05--09--06--11--07--13--08--15--09
    //       |       |       |       |       |
    //       01  00  03  01  05  02  07  03  08
    //       |       |       |       |       |
    //       00--00--01--02--02--04--03--06--04
    //
    // mask for the source grid (cells with "x" are masked)
    //
    //       +---+---+---+---+
    //       | x | x | x |   |
    //       +---+---+---+---+
    //       | x | x |   |   |
    //       +---+---+---+---+
    //       | x |   |   |   |
    //       +---+---+---+---+
    //       |   |   |   |   |
    //       +---+---+---+---+

    double coordinates_x[] = {0,1,2,3,4};
    double coordinates_y[] = {0,1,2,3,4};
    size_t const num_cells[] = {4,4};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;

    cdo_generate_grid_reg2d(srcGrid, coordinates_x, coordinates_y, num_cells);
  }

  {
    // the global target grid is a 3x3 grid:
    //
    //       12--21--13--22--14--23--15
    //       |       |       |       |
    //       15  06  17  07  19  08  20
    //       |       |       |       |
    //       08--14--09--16--10--18--11
    //       |       |       |       |
    //       08  03  10  04  12  05  13
    //       |       |       |       |
    //       04--07--05--09--06--11--07
    //       |       |       |       |
    //       01  00  03  01  05  02  06
    //       |       |       |       |
    //       00--00--01--02--02--04--03

    double coordinates_x[] = {0.5,1.5,2.5,3.5};
    double coordinates_y[] = {0.5,1.5,2.5,3.5};
    size_t const num_cells[2] = {3,3};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;
    
    cdo_generate_grid_reg2d(tgtGrid, coordinates_x, coordinates_y, num_cells);
  }

  Varray<double> srcField(srcGrid.numCells);
  Varray<double> tgtField(tgtGrid.numCells);

  int global_src_mask [] = {1,1,1,1,
                            0,1,1,1,
                            0,0,1,1,
                            0,0,0,1};
  for (size_t i = 0; i < srcGrid.numCells; ++i) srcField[i] = global_src_mask[i] ? i : missval;

  NormOpt normalisation[2] = {NormOpt::DESTAREA, NormOpt::FRACAREA};

  for (int norm_idx = 0; norm_idx <= 1; ++norm_idx)
    {
      gridcell_search_create(remapSearch.gcs, srcGrid.numCells, srcGrid.numCorners, srcGrid.cellCornersLon, srcGrid.cellCornersLat);

      remap_conserv(normalisation[norm_idx], remapSearch, srcField, tgtField, missval);

      gridcell_search_delete(remapSearch.gcs);

      double tgtFieldRef[2][9] = {{1.5,3.5,4.5,1.25,5.25,8.5,-1,2.5,9.0},
                                  {2.0,3.5,4.5,5.0,7.0,8.5,-1,10.0,12.0}};

      for (size_t i = 0; i < tgtGrid.numCells; ++i)
        {
          if (fabs(tgtFieldRef[norm_idx][i] - tgtField[i]) > 1e-3)
            printf("%s: wrong interpolation result\n", __func__);
        }
    }
}

static void test6()
{
  const double missval = -1.0;
  RemapSearch remapSearch;

  auto &srcGrid = remapSearch.srcGrid;
  auto &tgtGrid = remapSearch.tgtGrid;

  {
    // the global source grid is a 3x3 grid:
    //
    //       12--21--13--22--14--23--15
    //       |       |       |       |
    //       15  06  17  07  19  08  20
    //       |       |       |       |
    //       08--14--09--16--10--18--11
    //       |       |       |       |
    //       08  03  10  04  12  05  13
    //       |       |       |       |
    //       04--07--05--09--06--11--07
    //       |       |       |       |
    //       01  00  03  01  05  02  06
    //       |       |       |       |
    //       00--00--01--02--02--04--03

    double coordinates_x[] = {-1.5,-0.5,0.5,1.5};
    double coordinates_y[] = {-1.5,-0.5,0.5,1.5};
    size_t const num_cells[2] = {3,3};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;

    cdo_generate_grid_reg2d(srcGrid, coordinates_x, coordinates_y, num_cells);
  }

  {
    // the global target grid is a 2x2 grid:
    //
    //       06--10--07--11--08
    //       |       |       |
    //       06  02  08  03  09
    //       |       |       |
    //       03--05--04--07--05
    //       |       |       |
    //       01  00  03  01  04
    //       |       |       |
    //       00--00--01--02--02

    double coordinates_x[] = {-0.5,0,0.5};
    double coordinates_y[] = {-0.5,0,0.5};
    size_t const num_cells[2] = {2,2};
    for (size_t i = 0; i <= num_cells[0]; ++i) coordinates_x[i] *= DEG2RAD;
    for (size_t i = 0; i <= num_cells[1]; ++i) coordinates_y[i] *= DEG2RAD;
    
    cdo_generate_grid_reg2d(tgtGrid, coordinates_x, coordinates_y, num_cells);
  }

  //grid_print(srcGrid);
  //grid_print(tgtGrid);
  Varray<double> srcField(srcGrid.numCells);
  Varray<double> tgtField(tgtGrid.numCells);

  constexpr int NUM_TESTS = 4;

  double global_field_data[NUM_TESTS][9] = {
          {1,1,1, 1,1,1, 1,1,1},
          {0,0,0, 1,1,1, 2,2,2},
          {0,1,2, 0,1,2, 0,1,2},
          {0,1,2, 1,2,3, 2,3,4}};

  for (size_t t = 0; t < NUM_TESTS; ++t)
    {
      for (size_t i = 0; i < srcGrid.numCells; ++i) srcField[i] = global_field_data[t][i];

      gridcell_search_create(remapSearch.gcs, srcGrid.numCells, srcGrid.numCorners, srcGrid.cellCornersLon, srcGrid.cellCornersLat);

      remap_conserv2(NormOpt::DESTAREA, remapSearch, srcField, tgtField, missval);

      gridcell_search_delete(remapSearch.gcs);

      {
        double tgtFieldRef[NUM_TESTS][4] = {
          {1,1,
           1,1},
          {0.75,0.75,
           1.25,1.25},
          {0.75,1.25,
           0.75,1.25},
          {1.5,2.0,
           2.0,2.5}};

        for (size_t i = 0; i < tgtGrid.numCells; ++i) printf("%zu %g %g\n", i + 1, tgtFieldRef[t][i], tgtField[i]);
        for (size_t i = 0; i < tgtGrid.numCells; ++i)
          {
            if (fabs(tgtFieldRef[t][i] - tgtField[i]) > 1e-3)
              printf("%s: wrong interpolation result\n", __func__);
          }
      }
    }
}

int main(void)
{
  // 1st order conservativ remapping
  test1();
  test3();
  test4();

  // 2nd order conservativ remapping
  test6();

  return 0;
}
